/* $Id: cbc.c,v 1.2 2002/01/24 23:21:25 levitte Exp $ */

/* ====================================================================
 * Copyright (c) 1999 Levitte Programming.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *      This product includes software developed by Levitte Programming
 *      and its contributors.
 *
 * 4. The names "RLCRYPTO" and "Levitte Programming" or the names of its
 *    contributors must not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *    For written permission, please contact levitte@lp.se.
 *
 * THIS SOFTWARE IS PROVIDED BY LEVITTE PROGRAMMING AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL LEVITTE
 * PROGRAMMING OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

#include <stdlib.h>
#include <string.h>
#include "../algo_int.h"

struct algorithm_context_st
{
  unsigned char iv[64];
  size_t iv_len;
};

static size_t cbc_ctx_size(void)
{
  return sizeof(ALGO_CTX);
}

static ALGO_CTX *cbc_alloc(void)
{
  return (ALGO_CTX *)malloc(cbc_ctx_size());
}

static void cbc_free(ALGO_CTX *ctx)
{
  free(ctx);
}

static int cbc_init(ALGO_CTX *ctx, const unsigned char *ivec, size_t iveclen)
{
  memset(ctx->iv, 0, sizeof(ctx->iv));
  memcpy(ctx->iv, ivec, iveclen);
  if (iveclen > 64)
    return ALGO_E_TOO_LARGE_BLOCK;
  ctx->iv_len = iveclen;
  return ALGO_E_OK;
}

static void cbc_clear(ALGO_CTX *ctx)
{
  memset(ctx, 0, cbc_ctx_size());
}

static int cbc_enc_dec(ALGO_CTX *ctx,
		       const unsigned char *input, unsigned char *output, size_t length,
		       int encrypt,
		       int (*sym_enc_dec)(ALGO_CTX *ctx,
					  const unsigned char *input,
					  unsigned char *output, size_t length,
					  int encrypt),
		       ALGO_CTX *sym_data)
{
  int status = ALGO_E_OK;
  size_t n;
  unsigned char tmp[64];
  register size_t iv_len = ctx->iv_len;

  if (length % iv_len != 0)
    return ALGO_E_NOT_EXACT_BLOCKS;

  while (length)
    {
      if (encrypt)
	{
	  for(n=0; n < iv_len; ++n)
	    tmp[n] = input[n] ^ ctx->iv[n];
	  status = sym_enc_dec(sym_data, tmp, output, iv_len, encrypt);
	  if (status != ALGO_E_OK)
	    return status;
	  memcpy(ctx->iv, output, iv_len);
	}
      else
	{
	  memcpy(tmp, input, iv_len);
	  status = sym_enc_dec(sym_data, input, output, iv_len, encrypt);
	  if (status != ALGO_E_OK)
	    return status;
	  for(n=0; n < iv_len; ++n)
	    output[n] ^= ctx->iv[n];
	  memcpy(ctx->iv, tmp, iv_len);
	}
      input += iv_len;
      output += iv_len;
      length -= iv_len;
    }
  return status;
}

static int cbc_cpy_ivec(ALGO_CTX *ctx, unsigned char *ivec, size_t ivecsize,
			size_t *iveclen)
{
  if (ivecsize < ctx->iv_len)
    return ALGO_E_INVALID_IVEC_SIZE;
  memcpy(ivec, ctx->iv, ctx->iv_len);
  *iveclen = ctx->iv_len;
  return ALGO_E_OK;
}

static const MODE_METHOD cbc_method =
  {
    "CBC",
    0,
    1,
    &cbc_ctx_size,
    &cbc_alloc,
    &cbc_free,
    &cbc_init,
    &cbc_clear,
    &cbc_enc_dec,
    &cbc_cpy_ivec
  };

const MODE_METHOD *CBC(void)
{
  return &cbc_method;
}
